/*
 * Sparse Matrix Module
 *
 * Placed in the public domain.
 * April 05, 2000
 * Neil Schemenauer <nas@python.ca>
 *
 */


/* SparseMatrix objects */

#include <stdio.h>
#include <assert.h>
#include "Python.h"
#include "sparse/spMatrix.h"
#include "sparse/spDefs.h"

/* #define NDEBUG */

static PyObject *ErrorObject;

typedef struct {
    PyObject_HEAD
    PyObject *element_attr; /* attributes */
    double *value;
    PyObject *matrix;
} SparseElementObject;

staticforward PyTypeObject SparseElement_Type;

#define SparseElementObject_Check(v)    ((v)->ob_type == &SparseElement_Type)

static PyObject *
newSparseElementObject(PyObject *matrix, double *val)
{
    SparseElementObject *self;
    self = PyObject_NEW(SparseElementObject, &SparseElement_Type);
    if (self == NULL)
        return NULL;
    self->element_attr = NULL;
    self->matrix = matrix;
    Py_INCREF(matrix);
    self->value = val;
    return (PyObject *)self;
}

/* SparseElement methods */

static void
SparseElement_dealloc(SparseElementObject *self)
{
    Py_XDECREF(self->matrix);
    Py_XDECREF(self->element_attr);
    PyMem_DEL(self);
}

PyObject *
SparseElement_get(SparseElementObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ""))
        return NULL;
    assert(self->value!=NULL);
    return Py_BuildValue("d", *(self->value));
}

PyObject *
SparseElement_set(SparseElementObject *self, PyObject *args)
{
    double n;
    if (!PyArg_ParseTuple(args, "d", &n))
        return NULL;
    assert(self->value!=NULL);
    *(self->value) = n;
    Py_INCREF(Py_None);
    return Py_None;
}

static PyMethodDef SparseElement_methods[] = {
    {"get", (PyCFunction)SparseElement_get, 1},
    {"set", (PyCFunction)SparseElement_set, 1},
    {NULL,      NULL}       /* sentinel */
};

static PyObject *
SparseElement_getattr(SparseElementObject *self, char *name)
{
    if (self->element_attr != NULL) {
        PyObject *v = PyDict_GetItemString(self->element_attr, name);
        if (v != NULL) {
            Py_INCREF(v);
            return v;
        }
    }
    return Py_FindMethod(SparseElement_methods, (PyObject *)self, name);
}

static int
SparseElement_setattr(SparseElementObject *self, char *name, PyObject *v)
{
    if (self->element_attr == NULL) {
        self->element_attr = PyDict_New();
        if (self->element_attr == NULL)
            return -1;
    }
    if (v == NULL) {
        int rv = PyDict_DelItemString(self->element_attr, name);
        if (rv < 0)
            PyErr_SetString(PyExc_AttributeError,
                    "delete non-existing SparseElement attribute");
        return rv;
    }
    else
        return PyDict_SetItemString(self->element_attr, name, v);
}

static PyObject *
SparseElement_repr(SparseElementObject *v)
{
    char buf[200];
    assert(v!=NULL);
    sprintf(buf, "<MatrixElement at 0x%x=%.12g>", (int)v, *(v->value)); 
    return PyString_FromString(buf);
}


staticforward PyTypeObject SparseElement_Type = {
/*  PyObject_HEAD_INIT(&PyType_Type) */
    PyObject_HEAD_INIT(0)
    0,          /*ob_size*/
    "SparseElement",            /*tp_name*/
    sizeof(SparseElementObject),    /*tp_basicsize*/
    0,          /*tp_itemsize*/
    /* methods */
    (destructor)SparseElement_dealloc, /*tp_dealloc*/
    0,      /*tp_print*/
    (getattrfunc)SparseElement_getattr, /*tp_getattr*/
    (setattrfunc)SparseElement_setattr, /*tp_setattr*/
    0,          /*tp_compare*/
    (reprfunc)SparseElement_repr,   /*tp_repr*/
    0,          /*tp_as_number*/
    0,          /*tp_as_sequence*/
    0,          /*tp_as_mapping*/
    0,          /*tp_hash*/
};


/* --------------------------------------------------------------------- */

typedef struct {
    PyObject_HEAD
    PyObject *sp_attr; /* attributes */
    char *matrix;
} SparseMatrixObject;


staticforward PyTypeObject SparseMatrix_Type;

#define SparseMatrixObject_Check(v) ((v)->ob_type == &SparseMatrix_Type)

static PyObject *
newSparseMatrixObject(PyObject *args)
{
    int size;
    int result;
    SparseMatrixObject *p = PyObject_NEW(SparseMatrixObject, &SparseMatrix_Type);
    if (p == NULL) {
        return NULL;
    }
    size = 0;
    if (!PyArg_ParseTuple(args, "|i", &size)) {
        return NULL;
    }
    p->sp_attr = NULL;
    p->matrix = spCreate(size, 0, &result);
    assert(p->matrix!=NULL);
    if (result != 0) {
        PyErr_SetString(PyExc_MemoryError, "cannot create sparse matrix");
        return NULL;
    }
    return (PyObject *)p;
}

/* SpareMatrix methods */

static void
SparseMatrix_dealloc(SparseMatrixObject *self)
{
    assert(self->matrix!=NULL);
    spDestroy(self->matrix);
    Py_XDECREF(self->sp_attr);
    PyMem_DEL(self);
}

static PyObject *
SparseMatrix_clear(SparseMatrixObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, "")) {
        return NULL;
    }
    assert(self->matrix!=NULL);
    spClear(self->matrix);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
SparseMatrix_get_element(SparseMatrixObject *self, PyObject *args)
{
    int i, j;
    double *v;
    if (!PyArg_ParseTuple(args, "ii", &i, &j)) {
        return NULL;
    }
    if ( i<0 || j<0 ) {
        PyErr_SetString(PyExc_ValueError, "index less than 0");
        return NULL;
    }
    assert(self->matrix!=NULL);
    v = spGetElement(self->matrix, i+1, j+1);
    if (v == NULL) {
        PyErr_SetString(PyExc_MemoryError, "cannot get element");
        return NULL;
    }
    return newSparseElementObject((PyObject *)self, v);
}

static PyObject *
SparseMatrix_set_element(SparseMatrixObject *self, PyObject *args)
{
    int i, j;
    double *v, n;
    if (!PyArg_ParseTuple(args, "iid", &i, &j, &n)) {
        return NULL;
    }
    if ( i<0 || j<0 ) {
        PyErr_SetString(PyExc_ValueError, "index less than 0");
        return NULL;
    }
    assert(self->matrix!=NULL);
    v = spGetElement(self->matrix, i+1, j+1);
    if (v == NULL) {
        PyErr_SetString(PyExc_MemoryError, "cannot get element");
        return NULL;
    }
    *v = n;
    Py_INCREF(Py_None);
    return Py_None;
}



static PyObject *
SparseMatrix_condition(SparseMatrixObject *self, PyObject *args)
{
    int result;
    double norm, condition;
    norm = 1.0;
    if (!PyArg_ParseTuple(args, "|d", &norm)) {
        return NULL;
    }
    if (!IS_FACTORED((MatrixPtr)self->matrix)) {
        PyErr_SetString(ErrorObject, "matrix not factored");
        return NULL;
    }
    assert(self->matrix!=NULL);
    condition = spCondition(self->matrix, norm, &result);
    switch (result) {
        case spOKAY:
            return Py_BuildValue("d", condition);
        case spSINGULAR:
            PyErr_SetString(ErrorObject, "singular matrix");
            return NULL;
        case spNO_MEMORY:
            PyErr_SetString(PyExc_MemoryError, "no memory");
            return NULL;
        default:
            PyErr_SetString(ErrorObject, "unknown error");
            return NULL;
    }
}


static PyObject *
SparseMatrix_factor(SparseMatrixObject *self, PyObject *args)
{
    int result;
    if (!PyArg_ParseTuple(args, "")) {
        return NULL;
    }
    if (IS_FACTORED((MatrixPtr)self->matrix)) {
        Py_INCREF(Py_None);
        return Py_None;
    }
    assert(self->matrix!=NULL);
    result = spFactor(self->matrix);
    switch (result) {
        case spOKAY:
            Py_INCREF(Py_None);
            return Py_None;
        case spSINGULAR:
            PyErr_SetString(ErrorObject, "singular matrix");
            return NULL;
        case spNO_MEMORY:
            PyErr_SetString(PyExc_MemoryError, "no memory");
            return NULL;
        default:
            PyErr_SetString(ErrorObject, "unknown error");
            return NULL;
    }
}

static int
sequence_to_double(PyObject *seq, double **ret)
{
    PyObject *o;
    double *d;
    int length;
    int i;

    if (!PySequence_Check(seq)) {
        PyErr_SetString(PyExc_ValueError, "sequence expected");
        return -1;
    }
    assert(seq!=NULL);
    length = PySequence_Length(seq);
    if (length < 0) {
        PyErr_SetString(PyExc_ValueError, "cannot find length of sequence");
        return -1;
    }
    d = malloc(sizeof(double)*length);
    if (d == NULL) {
        PyErr_SetString(PyExc_MemoryError, "no memory");
        return -1;
    }
    for (i=0; i<length; i++) {
        o = PySequence_GetItem(seq, i);
        if (o == NULL) {
            PyErr_SetString(PyExc_ValueError, "can't get element");
            free(d);
            return -1;
        }
        o = PyNumber_Float(o);
        if ( o==NULL || (!PyFloat_Check(o))) {
            PyErr_SetString(PyExc_ValueError, "element not a number");
            free(d);
            return -1;
        }
        d[i] = PyFloat_AsDouble(o);
    }
    *ret = d;
    return length;
}

PyObject *
double_to_list(double *d, int length) {
    int i;
    PyObject *l = PyList_New(length);
    assert(l!=NULL);
    for (i=0; i<length; i++) {
        PyList_SetItem(l, i, Py_BuildValue("d", d[i]));
    }
    return l;
}

static PyObject *
SparseMatrix_test(SparseMatrixObject *self, PyObject *args)
{
    double *rhs;
    int length;
    PyObject *seq;
    if (!PyArg_ParseTuple(args, "O", &seq)) {
        return NULL;
    }
    if ((length=sequence_to_double(seq, &rhs)) == -1) {
        return NULL;
    }
    return double_to_list(rhs, length);
}

static PyObject *
SparseMatrix_order_and_factor(SparseMatrixObject *self, PyObject *args)
{
    double *rhs, threshold, abs_threshold;
    int diag_pivot;
    int result;
    PyObject *seq;
    int length;
    
    threshold = 0.0;
    abs_threshold = 1.0e-15;
    diag_pivot = 0;
    if (!PyArg_ParseTuple(args, "O|ddi", &seq, &threshold, &abs_threshold, &diag_pivot)) {
        return NULL;
    }
    if ((length=sequence_to_double(seq, &rhs)) == -1) {
        return NULL;
    }
    assert(self->matrix!=NULL);
    assert(rhs!=NULL);
    if (length < spGetSize(self->matrix, 1)) {
        PyErr_SetString(PyExc_ValueError, "sequence is too short");
        return NULL;
    }
    if (IS_FACTORED((MatrixPtr)self->matrix)) {
        Py_INCREF(Py_None);
        return Py_None;
    }
    result = spOrderAndFactor(self->matrix, rhs, threshold, abs_threshold, diag_pivot);
    assert(rhs!=NULL);
    free(rhs);
    switch (result) {
        case spOKAY:
            Py_INCREF(Py_None);
            return Py_None;
        case spSINGULAR:
            PyErr_SetString(ErrorObject, "singular matrix");
            return NULL;
        case spNO_MEMORY:
            PyErr_SetString(PyExc_MemoryError, "no memory");
            return NULL;
        default:
            PyErr_SetString(ErrorObject, "unknown error");
            return NULL;
    }
}


static PyObject *
SparseMatrix_solve(SparseMatrixObject *self, PyObject *args)
{
    double *rhs;
    PyObject *seq;
    int length;
    
    if (!PyArg_ParseTuple(args, "O", &seq)) {
        return NULL;
    }
    assert(seq!=NULL);
    if ((length=sequence_to_double(seq, &rhs)) == -1) {
        return NULL;
    }
    assert(self->matrix!=NULL);
    assert(rhs!=NULL);
    if (length < spGetSize(self->matrix, 1)) {
        PyErr_SetString(PyExc_ValueError, "sequence is too short");
        return NULL;
    }
    if (!IS_FACTORED((MatrixPtr)self->matrix)) {
        PyErr_SetString(ErrorObject, "matrix is not factored");
        return NULL;
    }
    spSolve(self->matrix, rhs, rhs);
    seq = double_to_list(rhs, length);
    assert(seq!=NULL);
    assert(rhs!=NULL);
    free(rhs);
    return seq;
}

static PyObject *
SparseMatrix_solve_transposed(SparseMatrixObject *self, PyObject *args)
{
    double *rhs;
    PyObject *seq;
    int length;
    
    if (!PyArg_ParseTuple(args, "O", &seq)) {
        return NULL;
    }
    assert(seq!=NULL);
    if ((length=sequence_to_double(seq, &rhs)) == -1) {
        return NULL;
    }
    assert(self->matrix!=NULL);
    assert(rhs!=NULL);
    if (length < spGetSize(self->matrix, 1)) {
        PyErr_SetString(PyExc_ValueError, "sequence is too short");
        return NULL;
    }
    if (!IS_FACTORED((MatrixPtr)self->matrix)) {
        PyErr_SetString(ErrorObject, "matrix is not factored");
        return NULL;
    }
    spSolve(self->matrix, rhs, rhs);
    seq = double_to_list(rhs, length);
    assert(seq!=NULL);
    assert(rhs!=NULL);
    free(rhs);
    return seq;
}

static PyObject *
SparseMatrix_multiply(SparseMatrixObject *self, PyObject *args)
{
    double *rhs, *solution;
    PyObject *seq;
    int length;
    
    if (!PyArg_ParseTuple(args, "O", &seq)) {
        return NULL;
    }
    assert(self->matrix!=NULL);
    if (IS_FACTORED((MatrixPtr)self->matrix)) {
        PyErr_SetString(ErrorObject, "matrix is factored");
        return NULL;
    }
    assert(seq!=NULL);
    if ((length=sequence_to_double(seq, &solution)) == -1) {
        return NULL;
    }
    if (length < spGetSize(self->matrix, 1)) {
        PyErr_SetString(PyExc_ValueError, "sequence is too short");
        return NULL;
    }
    rhs = malloc(sizeof(double)*length);
    if (rhs==NULL) {
        PyErr_SetString(PyExc_MemoryError, "out of memory");
        return NULL;
    }
    spMultiply(self->matrix, rhs, solution);
    seq = double_to_list(rhs, length);
    assert(seq!=NULL);
    assert(rhs!=NULL);
    free(rhs);
    assert(solution!=NULL);
    free(solution);
    return seq;
}


PyObject *
SparseMatrix_write(SparseMatrixObject *self, PyObject *args)
{
    PyObject *f;
    int reorder, data, header;
    FILE *fp;

    reorder = 1;
    data = 1;
    header = 0;
    if (!PyArg_ParseTuple(args, "O|iii", &f, &reorder, &data, &header)) {
        return NULL;
    }
    assert(f!=NULL);
    fp = PyFile_AsFile(f);
    assert(fp!=NULL);
    assert(self->matrix!=NULL);
    spPrint(self->matrix, fp, reorder, data, header);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
SparseMatrix_strip_fills(SparseMatrixObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, "")) {
        return NULL;
    }
    assert(self->matrix!=NULL);
    spStripFills(self->matrix);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
SparseMatrix_where_singular(SparseMatrixObject *self, PyObject *args)
{
    int i, j;
    if (!PyArg_ParseTuple(args, "")) {
        return NULL;
    }
    assert(self->matrix!=NULL);
    spWhereSingular(self->matrix, &i, &j);
    return Py_BuildValue("(ii)", i, j);
}

static PyObject *
SparseMatrix_fillin_count(SparseMatrixObject *self, PyObject *args)
{
    int count;
    if (!PyArg_ParseTuple(args, "")) {
        return NULL;
    }
    assert(self->matrix!=NULL);
    count = spFillinCount(self->matrix);
    return Py_BuildValue("i", count);
}

static PyObject *
SparseMatrix_largest_element(SparseMatrixObject *self, PyObject *args)
{
    double size;
    if (!PyArg_ParseTuple(args, "")) {
        return NULL;
    }
    assert(self->matrix!=NULL);
    size = spLargestElement(self->matrix);
    return Py_BuildValue("d", size);
}

static PyObject *
SparseMatrix_norm(SparseMatrixObject *self, PyObject *args)
{
    double norm;
    if (!PyArg_ParseTuple(args, "")) {
        return NULL;
    }
    if (IS_FACTORED((MatrixPtr)self->matrix)) {
        PyErr_SetString(ErrorObject, "matrix factored");
        return NULL;
    }
    assert(self->matrix!=NULL);
    norm = spNorm(self->matrix);
    return Py_BuildValue("d", norm);
}

static PyObject *
SparseMatrix_roundoff(SparseMatrixObject *self, PyObject *args)
{
    double result;
    double rho;
    if (!PyArg_ParseTuple(args, "d", &rho)) {
        return NULL;
    }
    assert(self->matrix!=NULL);
    result = spRoundoff(self->matrix, rho);
    return Py_BuildValue("d", result);
}

static PyObject *
SparseMatrix_determinant(SparseMatrixObject *self, PyObject *args)
{
    int exponent;
    double result;
    exponent = 1;
    if (!PyArg_ParseTuple(args, "")) {
        return NULL;
    }
    if (!IS_FACTORED((MatrixPtr)self->matrix)) {
        PyErr_SetString(ErrorObject, "matrix not factored");
        return NULL;
    }
    assert(self->matrix!=NULL);
    spDeterminant(self->matrix, &exponent, &result);
    return Py_BuildValue("id", exponent, result);
}



static PyMethodDef sp_methods[] = {
    {"clear",   (PyCFunction)SparseMatrix_clear,    1},
    {"get_element", (PyCFunction)SparseMatrix_get_element,  1},
    {"set_element", (PyCFunction)SparseMatrix_set_element,  1},
    {"factor",  (PyCFunction)SparseMatrix_factor,   1},
    {"order_and_factor",    (PyCFunction)SparseMatrix_order_and_factor, 1},
    {"solve",   (PyCFunction)SparseMatrix_solve,    1},
    {"solve_transposed",    (PyCFunction)SparseMatrix_solve_transposed, 1},
    {"multiply",    (PyCFunction)SparseMatrix_multiply, 1},
    {"condition",   (PyCFunction)SparseMatrix_condition,    1},
    {"write",   (PyCFunction)SparseMatrix_write,    1},
    {"strip_fills", (PyCFunction)SparseMatrix_strip_fills,  1},
    {"where_singular",  (PyCFunction)SparseMatrix_where_singular,   1},
    {"fillin_count",    (PyCFunction)SparseMatrix_fillin_count, 1},
    {"largest_element", (PyCFunction)SparseMatrix_largest_element,  1},
    {"norm",    (PyCFunction)SparseMatrix_norm, 1},
    {"roundoff",    (PyCFunction)SparseMatrix_roundoff, 1},
    {"determinant", (PyCFunction)SparseMatrix_determinant,  1},
    {"test",    (PyCFunction)SparseMatrix_test, 1},
    {NULL,      NULL}       /* sentinel */
};

static PyObject *
SparseMatrix_getattr(SparseMatrixObject *self, char *name)
{
    if (self->sp_attr != NULL) {
        PyObject *v = PyDict_GetItemString(self->sp_attr, name);
        if (v != NULL) {
            Py_INCREF(v);
            return v;
        }
    }
    return Py_FindMethod(sp_methods, (PyObject *)self, name);
}

static int
SparseMatrix_setattr(SparseMatrixObject *self, char *name, PyObject *v)
{
    if (self->sp_attr == NULL) {
        self->sp_attr = PyDict_New();
        if (self->sp_attr == NULL)
            return -1;
    }
    if (v == NULL) {
        int rv = PyDict_DelItemString(self->sp_attr, name);
        if (rv < 0)
            PyErr_SetString(PyExc_AttributeError,
                    "delete non-existing SparseMatrix attribute");
        return rv;
    }
    else
        return PyDict_SetItemString(self->sp_attr, name, v);
}

staticforward PyTypeObject SparseMatrix_Type = {
/*  PyObject_HEAD_INIT(&PyType_Type) */
    PyObject_HEAD_INIT(0)
    0,          /*ob_size*/
    "SparseMatrix",     /*tp_name*/
    sizeof(SparseMatrixObject), /*tp_basicsize*/
    0,          /*tp_itemsize*/
    /* methods */
    (destructor)SparseMatrix_dealloc, /*tp_dealloc*/
    0,      /*tp_print*/
    (getattrfunc)SparseMatrix_getattr, /*tp_getattr*/
    (setattrfunc)SparseMatrix_setattr, /*tp_setattr*/
    0,          /*tp_compare*/
    0,          /*tp_repr*/
    0,          /*tp_as_number*/
    0,          /*tp_as_sequence*/
    0,          /*tp_as_mapping*/
    0,          /*tp_hash*/
};

/* --------------------------------------------------------------------- */

/* new SparseMatrix object */

static PyObject *
sparse_new(PyObject *self, PyObject *args)
{
    PyObject *rv;
    rv = newSparseMatrixObject(args);
    if ( rv == NULL )
        return NULL;
    return rv;
}

/* List of functions defined in the module */

static PyMethodDef sparse_methods[] = {
    {"sparse",  sparse_new, 1},
    {NULL,      NULL}       /* sentinel */
};


/* Initialization function for the module (*must* be called initsparse) */

void
initsparse(void)
{
    PyObject *m, *d;

    /* Create the module and add the functions */
    SparseMatrix_Type.ob_type = &PyType_Type;
    m = Py_InitModule("sparse", sparse_methods);

    /* Add some symbolic constants to the module */
    d = PyModule_GetDict(m);
    ErrorObject = PyErr_NewException("sparse.error", NULL, NULL);
    PyDict_SetItemString(d, "error", ErrorObject);
}
